import 'package:flutter/material.dart';

typedef _QueryFuture<T> = Future<T> Function();

class _CacheItem<T> {
  final List<Function> subscribers;
  T? data;

  _CacheItem({
    required this.subscribers,
  });

  void addData(T t) {
    data = t;
  }

  void addSubscriber(Function func) {
    subscribers.add(func);
  }

  List<Function> getSubscribers() {
    return subscribers;
  }

  void removeSubscribers() {
    subscribers.clear();
  }
}

class _Cache {
  Map<String, _CacheItem> _cache = {};

  void addSubscriber(String key, Function func) {
    if (_cache.containsKey(key)) {
      _cache[key]!.addSubscriber(func);
    } else {
      _cache[key] = _CacheItem(subscribers: [func]);
    }
  }

  List<Function> getSubscribers(String key) {
    if (_cache.containsKey(key)) {
      return _cache[key]!.getSubscribers();
    } else {
      return [];
    }
  }

  void removeSubscribers(String key) {
    if (_cache.containsKey(key)) {
      _cache[key]!.removeSubscribers();
    }
  }

  void notifySubscribers(String key) {
    for (var subscriber in this.getSubscribers(key)) {
      try {
        subscriber();
      } catch (e) {}
    }
  }

  void addData<T>(String key, T t) {
    if (_cache.containsKey(key)) {
      _cache[key]!.addData(t);
    }
  }

  T? getData<T>(String key) {
    if (_cache.containsKey(key)) {
      return _cache[key]!.data;
    }
    return null;
  }

  void reset() {
    _cache = {};
  }
}

var cache = _Cache();

class QueryResponse<T> {
  T? data;
  bool loading;
  dynamic error;

  QueryResponse({
    required this.data,
    required this.loading,
    required this.error,
  });
}

typedef _QueryBuilder = Widget Function(
  BuildContext context,
  dynamic data,
  bool loading,
  dynamic error,
);

class Query<T> extends StatefulWidget {
  final String cacheKey;
  final _QueryFuture<T> future;
  final _QueryBuilder builder;

  Query({
    required this.cacheKey,
    required this.builder,
    required this.future,
  });

  @override
  _QueryState<T> createState() => _QueryState<T>();
}

class _QueryState<T> extends State<Query<T>> {
  late Stream<QueryResponse<T>> stream;
  Stream<QueryResponse<T>> getStream<T>(
    Function cb,
  ) async* {
    try {
      T? data = cache.getData<T>(widget.cacheKey);
      yield QueryResponse(
        data: data,
        loading: true,
        error: null,
      );
      T response = await cb();
      cache.addData(widget.cacheKey, response);
      yield QueryResponse<T>(
        data: response,
        loading: false,
        error: null,
      );
    } catch (e) {
      print(e);
      yield QueryResponse<T>(
        data: null,
        loading: false,
        error: e,
      );
    }
  }

  @override
  void initState() {
    stream = getStream(widget.future);
    cache.addSubscriber(widget.cacheKey, _rebuild);
    super.initState();
  }

  void _rebuild() {
    setState(() {
      stream = getStream(widget.future);
    });
  }

  @override
  Widget build(BuildContext context) {
    return StreamBuilder(
      initialData: QueryResponse(data: null, loading: true, error: null),
      stream: stream,
      builder: (BuildContext context, AsyncSnapshot<QueryResponse> snapshot) {
        QueryResponse data = snapshot.data!;
        return this.widget.builder(
              context,
              data.data,
              data.loading,
              data.error,
            );
      },
    );
  }
}

void refetchQuery(String key) {
  cache.notifySubscribers(key);
}

void removeQuery(String key) {
  cache.removeSubscribers(key);
}

void resetStore() {
  cache.reset();
}
